# -*- coding: utf-8 -*-
"""
This module encapsulates all shell callable functions of nidaba.
"""

from __future__ import absolute_import, print_function

from inspect import getcallargs, getdoc
from gunicorn.six import iteritems
from itertools import cycle
from pprint import pprint
from flask import Flask

from nidaba.nidaba import NetworkSimpleBatch, SimpleBatch

import uuid
import shutil
import os.path
import sys
import click
import stevedore
import logging
import gunicorn.app.base

spinner = cycle([u'⣾', u'⣽', u'⣻', u'⢿', u'⡿', u'⣟', u'⣯', u'⣷'])

def spin(msg):
    click.echo(u'\r\033[?25l{}\t\t{}'.format(msg, next(spinner)), nl=False)


@click.group(epilog='This nidaba may or may not have Super Cow Powers')
@click.version_option()
def main():
    """
    Sends jobs to nidaba and retrieves their status.
    """


def int_float_bool_or_str(s):
    """
    A small helper function intended to coerce an input string to types in the
    order int -> float -> bool -> unicode -> input.

    Args:
        s (unicode):

    Returns:
        int or float or unicode or original input type: Input variable coerced
        to the highest data type in the ordering.
    """

    try:
        return int(s)
    except ValueError:
        try:
            return float(s)
        except ValueError:
            if s in ['True', 'true']:
                return True
            elif s in ['False', 'false']:
                return False
            try:
                return unicode(s)
            except UnicodeDecodeError:
                return s


def help_tasks(ctx, param, value):
    if not value or ctx.resilient_parsing:
        return
    from nidaba import celery
    t = celery.app.tasks
    hidden = ['util']
    last = u''
    docs = u''
    for k in sorted(t.keys()):
        hier = k.split('.')
        if hier[0] == 'nidaba':
            if hier[1] in hidden:
                continue
            elif hier[1] != last:
                docs += '{}\n{}\n\n'.format(hier[1].title(), len(hier[1]) *
                                            '-')
                last = hier[1]
            docs += '{}\n{}\n\n{}'.format(hier[-1], len(hier[-1]) * '~',
                                          getdoc(t[k].run).decode('utf-8').partition('Returns:')[0])
    click.echo_via_pager(docs)
    ctx.exit()


def validate_definition(ctx, param, value):
    """
    Validates all task definitions of a group and returns them as a list.
    """
    definitions = []
    for alg in value:
        (task, _, params) = alg.partition(u':')
        configurations = []
        for conf in params.split(u';'):
            kwargs = {}
            for arg in conf.split(u','):
                if '=' in arg:
                    k, v = arg.split('=')
                    kwargs[k] = int_float_bool_or_str(v)
                elif arg:
                    raise click.BadParameter('Positional arguments are deprecated!')
            configurations.append(kwargs)
        definitions.append([task, configurations])
    return definitions


def move_to_storage(batch, kwargs):
    """
    Takes as dictionary of kwargs and moves the suffix of all keys starting
    with the string 'file:' to the storage medium, prepending a unique
    identifier. The path components are rewritten in storage tuple form and the
    modified dictionary is returned.

    It is assumed that the filestore is already created.
    """
    from nidaba import storage

    nkwargs = {}
    def do_move(batch, src):
        if isinstance(batch, NetworkSimpleBatch):
            dst = os.path.basename(src)
            def callback(monitor):
                spin(u'Uploading {}'.format(dst))
            batch.add_document(src, callback, auxiliary=True)
            click.secho(u'\b\u2713', fg='green', nl=False)
            click.echo('\033[?25h\n', nl=False)
        else:
            suffix = uuid.uuid4()
            dst = unicode(suffix) + '_' + os.path.basename(src)
            shutil.copy2(src, storage.get_abs_path(batch.id, dst))
        return (batch.id, dst)
    for k, v in kwargs.iteritems():
        if isinstance(v, basestring) and v.startswith('file:'):
            v = v.replace('file:', '', 1)
            nkwargs[k] = do_move(batch, v)
        else:
            nkwargs[k] = v
    return nkwargs


@main.command()
@click.option('-h', '--host', default=None, 
              help='Address of the API service. If none is given a local '
              'installation of nidaba will be invoked.')
@click.option('--binarize', '-b', multiple=True, callback=validate_definition,
              help='A configuration for a single binarization algorithm in '
              'the format algorithm:param1,param2;param1,param2;...')
@click.option('--segmentation', '-l', multiple=True,
              callback=validate_definition,
              help='A configuration for a single page segmentation algorithm in '
              'the format algorithm:param1,param2;param1,param2;...')
@click.option('--ocr', '-o', multiple=True, callback=validate_definition,
              help='A configuration for a single OCR engine in the format '
              'engine:param1,param2;param1,param2;...')
@click.option('--stats', '-s', multiple=True, callback=validate_definition,
              help='A configuration for a single post-OCR measure in the '
              'format measure:param1,param2;param1;param2...')
@click.option('--postprocessing', '-p', multiple=True,
              callback=validate_definition, help='A configuration for a '
              'single postprocessing task in the format '
              'task:param1,param2;param1;param1...')
@click.option('--output', '-f', multiple=True,
              callback=validate_definition, help='A configuration for a '
              'single output layer transformation in the format'
              'task:param1,param2;param1;param1...')
# @click.option('--willitblend', 'blend',  default=False, help='Blend all '
#              'output files into a single hOCR document.', is_flag=True)
@click.option('--grayscale', default=False, help='Skip grayscale '
              'conversion using the ITU-R 601-2 luma transform if the input '
              'documents are already in grayscale.', is_flag=True)
@click.option('--help-tasks', is_eager=True, is_flag=True, callback=help_tasks,
              help='Accesses the documentation of all tasks contained in '
              'nidaba itself and in configured plugins.')
@click.argument('files', type=click.Path(exists=True), nargs=-1, required=True)
def batch(files, host, binarize, ocr, segmentation, stats, postprocessing, output,
          grayscale, help_tasks):
    """
    Add a new job to the pipeline.
    """
   
    if host:
        batch = NetworkSimpleBatch(host)
        click.echo(u'Preparing filestore\t\t[', nl=False)
        try:
            batch.create_batch()
        except:
            click.secho(u'\u2717', fg='red', nl=False)
            click.echo(']')
            exit()
        click.secho(u'\u2713', fg='green', nl=False)
        click.echo(']')
        for doc in files:
            def callback(monitor):
                spin(u'Uploading {}'.format(doc))
            batch.add_document(doc, callback)
            click.secho(u'\b\u2713', fg='green', nl=False)
            click.echo('\033[?25h\n', nl=False)
    else:
        from nidaba import storage
        click.echo(u'Preparing filestore\t\t[', nl=False)
        try:
            batch = SimpleBatch()
        except:
            click.secho(u'\u2717', fg='red', nl=False)
            click.echo(']')
            exit()
        for doc in files:
            shutil.copy2(doc, storage.get_abs_path(batch.id, os.path.basename(doc)))
            batch.add_document((batch.id, os.path.basename(doc)))
        click.secho(u'\u2713', fg='green', nl=False)
        click.echo(']')
    click.echo(u'Building batch\t\t\t[', nl=False)
    if not grayscale:
        batch.add_task('img', 'rgb_to_gray')
    if binarize:
        for alg in binarize:
            for kwargs in alg[1]:
                kwargs = move_to_storage(batch, kwargs)
                batch.add_task('binarize', alg[0], **kwargs)
    if segmentation:
        for alg in segmentation:
            for kwargs in alg[1]:
                kwargs = move_to_storage(batch, kwargs)
                batch.add_task('segmentation', alg[0], **kwargs)
    if ocr:
        for alg in ocr:
            for kwargs in alg[1]:
                kwargs = move_to_storage(batch, kwargs)
                batch.add_task('ocr', alg[0], **kwargs)
    if stats:
        for alg in stats:
            for kwargs in alg[1]:
                kwargs = move_to_storage(batch, kwargs)
                batch.add_task('stats', alg[0], **kwargs)
    if postprocessing:
        for alg in postprocessing:
            for kwargs in alg[1]:
                kwargs = move_to_storage(batch, kwargs)
                batch.add_task('postprocessing', alg[0], **kwargs)
    if output:
        for alg in output:
            for kwargs in alg[1]:
                kwargs = move_to_storage(batch, kwargs)
                batch.add_task('output', alg[0], **kwargs)
    batch.run()
    click.secho(u'\u2713', fg='green', nl=False)
    click.echo(']')
    click.echo(batch.id)


@main.command()
def worker():
    """
    Starts a celery worker.
    """
    from nidaba import celery
    celery.app.worker_main(argv=sys.argv[:1])


@main.command()
@click.pass_context
@click.option('-b', '--bind', default='127.0.0.1:8080', 
              help='Address and port to bind the application worker to.')
@click.option('-w', '--workers', default=1, type=click.INT, 
              help='Number of request workers')
def api_server(ctx, **kwargs):
    """
    Starts the nidaba API server using gunicorn.
    """

    try:
        from nidaba import api
        from nidaba import web
    except IOError as e:
        if e.errno == 2:
            click.echo('No configuration file found at {}'.format(e.filename))
            ctx.exit()

    logging.basicConfig(level=logging.DEBUG)

    app = Flask('nidaba')
    app.register_blueprint(api.get_blueprint())
    app.register_blueprint(web.get_blueprint())

    class APIServer(gunicorn.app.base.BaseApplication):

        def __init__(self, app, options=None):
            self.options = options or {}
            self.application = app
            super(APIServer, self).__init__()

        def load_config(self):
            config = dict([(key, value) for key, value in iteritems(self.options)
                           if key in self.cfg.settings and value is not None])
            for key, value in iteritems(config):
                self.cfg.set(key.lower(), value)

        def load(self):
            return self.application

    APIServer(app, options=kwargs).run()


@main.command()
@click.pass_context
def config(ctx):
    """
    Displays the current configuration.
    """
    try:
        from nidaba.config import nidaba_cfg
    except IOError as e:
        if e.errno == 2:
            click.echo('No configuration file found at {}'.format(e.filename))
            ctx.exit()
    pprint(nidaba_cfg)


@main.command()
@click.pass_context
def plugins(ctx):
    """
    Displays available plugins and if they're enabled.
    """
    try:
        from nidaba.config import nidaba_cfg
    except IOError as e:
        if e.errno == 2:
            click.echo('plugins command only available for local installations.')
            ctx.exit()

    mgr = stevedore.ExtensionManager(namespace='nidaba.plugins')
    enabled = set(nidaba_cfg['plugins_load'].keys()).intersection(set(mgr.names()))
    disabled = set(mgr.names()) - enabled
    for plugin in mgr.names():
        click.echo(plugin, nl=False)
        if plugin in disabled:
            click.secho(u' (disabled)', fg='red')
        else:
            click.secho(u' (enabled)', fg='green')


@main.command()
@click.option('-v', '--verbose', count=True)
@click.option('-h', '--host', default=None, 
              help='Address of the API service. If none is given a local '
              'installation of nidaba will be invoked.')
@click.argument('job_id', nargs=1)
def status(verbose, host, job_id):
    """
    Diplays the status and results of jobs.
    """
    if host:
        batch = NetworkSimpleBatch(host, job_id)
    else:
        batch = SimpleBatch(job_id)

    state = batch.get_extended_state()

    click.secho('Status:', underline=True, nl=False)
    if not state:
        click.echo(' UNKNOWN')
        return

    bs = 'success'
    done = 0
    running = 0
    pending = 0
    failed = 0
    results = []
    errors = []
    for subtask in state.itervalues():
        if subtask['state'] == 'SUCCESS':
            done += 1
        elif subtask['state'] == 'RUNNING':
            running += 1
        elif subtask['state'] == 'PENDING':
            pending += 1
            if bs == 'success':
                bs = 'pending'
        elif subtask['state'] == 'FAILURE':
            failed += 1
            errors.append(subtask)
            bs = 'failed'

        if len(subtask['children']) == 0 and not subtask['housekeeping'] and subtask['result'] is not None:
            results.append((subtask['result'], subtask['root_document']))

    click.echo(' {}\n'.format(bs))
    click.echo('{}/{} tasks completed. {} running.\n'.format(done, len(state), running))
    click.secho('Output files:\n', underline=True)
    if results and host:
        for doc in results:
            click.echo(doc[1] + u' \u2192 ' + doc[0])
    elif results:
        from nidaba import storage
        for doc in results:
            output = click.format_filename(storage.get_abs_path(*doc[0]))
            click.echo(doc[1][1] + u' \u2192 ' + output)
    if errors:
        click.secho('\nErrors:\n', underline=True)
        for task in errors:
            click.echo('{0} ({1}): {2}'.format(task['task'][0],
                                               task['root_document'][1],
                                               task['errors'][-1]))



# -*- coding: utf-8 -*-
"""
nidaba.plugins.tesseract
~~~~~~~~~~~~~~~~~~~~~~~~

Plugin implementing an interface to tesseract

This plugin exposes tesseract's functionality as a task. It implements two ways
of calling tesseract, a direct method calling the tesseract executable and one
utilizing the C-API available from tesseract 3.02 and upwards.

The C-API requires a libtesseract shared object in the current library path and
training data in the configured tessdata directory:

.. code-block:: console

    # apt-get install libtesseract3 tesseract-ocr-$lang

Using the direct call method requires the tesseract binary installable by
executing:

.. code-block:: console

    # apt-get install tesseract-ocr

.. note::
    It is strongly encouraged to use the C-API whenever possible. It is
    supposedly stable while hOCR output file names change between tesseract
    versions.

.. note::
    Parameters in configuration files supersede command line parameters.
    Modular page segmentation utilizing zone files requires that the page
    segmentation mode may be set freely. Uncomment the line:

        tessedit_pageseg_mode 1

    in the default hocr configuration (in TESSDATA/configs/).

Configuration
~~~~~~~~~~~~~

implementation (default='capi')
    Selector for the call method. May either be `capi`, `direct` (tesseract
    hOCR output with .hocr extension), or `legacy` (tesseract hOCR output with
    .html extension).

tessdata (default='/usr/share/tesseract-ocr/')
    Path to load tesseract training data and configuration from. Has to be one
    directory level upwards from the actual tessdata directory
"""

from __future__ import unicode_literals, print_function, absolute_import

import subprocess
import StringIO
import ctypes
import os

from PIL import Image
from ctypes import POINTER
from distutils import spawn
from os.path import splitext
from celery.utils.log import get_task_logger

from nidaba.uzn import UZNWriter
from nidaba import storage
from nidaba.tei import TEIFacsimile
from nidaba.celery import app
from nidaba.tasks.helper import NidabaTask
from nidaba.nidabaexceptions import NidabaTesseractException
from nidaba.nidabaexceptions import NidabaPluginException


implementation = u'capi'
tessdata = u'/usr/share/tesseract-ocr/'

(RIL_BLOCK, RIL_PARA, RIL_TEXTLINE, RIL_WORD, RIL_SYMBOL) = map(ctypes.c_int,
                                                                xrange(5))

logger = get_task_logger(__name__)


class TessBaseAPI(ctypes.Structure):
    pass


class Pix(ctypes.Structure):
    pass


class TessResultIterator(ctypes.Structure):
    pass


class TessPageIterator(ctypes.Structure):
    pass


class TessResultRenderer(ctypes.Structure):
    pass


def setup(*args, **kwargs):
    if kwargs.get(u'implementation'):
        global implementation
        implementation = kwargs.get(u'implementation')
    if kwargs.get(u'tessdata'):
        global tessdata
        if isinstance(kwargs.get(u'tessdata'), list):
            tessdata = storage.get_abs_path(*kwargs.get(u'tessdata'))
        else:
            tessdata = kwargs.get(u'tessdata')
    if implementation == 'direct' and not spawn.find_executable('tesseract'):
        raise NidabaPluginException('No tesseract executable found')
    if implementation == 'capi':
        try:
            global tesseract, leptonica
            tesseract = ctypes.cdll.LoadLibrary('libtesseract.so.3')
            leptonica = ctypes.cdll.LoadLibrary('liblept.so')
        except:
            raise NidabaPluginException('Loading libtesseract/leptonica failed.')

        tesseract.TessBaseAPICreate.restype = POINTER(TessBaseAPI)

        tesseract.TessBaseAPIEnd.argtypes = [POINTER(TessBaseAPI)]
        tesseract.TessBaseAPIEnd.restype = None

        tesseract.TessBaseAPIDelete.argtypes = [POINTER(TessBaseAPI)]
        tesseract.TessBaseAPIDelete.restype = None

        tesseract.TessBaseAPIInit3.argtypes = [POINTER(TessBaseAPI),
                                               ctypes.c_char_p,
                                               ctypes.c_char_p]
        tesseract.TessBaseAPIInit3.restype = ctypes.c_int

        tesseract.TessBaseAPISetImage2.restype = None
        tesseract.TessBaseAPISetImage2.argtypes = [POINTER(TessBaseAPI),
                                                   POINTER(Pix)]

        tesseract.TessBaseAPIRecognize.argtypes = [POINTER(TessBaseAPI), POINTER(TessBaseAPI)]
        tesseract.TessBaseAPIRecognize.restype = ctypes.c_int

        tesseract.TessResultIteratorGetUTF8Text.restype = ctypes.c_char_p
        tesseract.TessResultIteratorGetUTF8Text.argtypes = [POINTER(TessResultIterator),
                                                            ctypes.c_int]

        tesseract.TessResultIteratorConfidence.argtypes = [POINTER(TessResultIterator),
                                                           ctypes.c_int]
        tesseract.TessResultIteratorConfidence.restype = ctypes.c_float

        tesseract.TessResultIteratorWordRecognitionLanguage.argtypes = [POINTER(TessResultIterator)]
        tesseract.TessResultIteratorWordRecognitionLanguage.restype = ctypes.c_char_p

        tesseract.TessVersion.restype = ctypes.c_char_p

        tesseract.TessBaseAPISetPageSegMode.argtypes = [POINTER(TessBaseAPI),
                                                        ctypes.c_int]
        tesseract.TessBaseAPISetPageSegMode.restype = None

        tesseract.TessBaseAPIProcessPages.argtypes = [POINTER(TessBaseAPI),
                                                      ctypes.c_char_p,
                                                      ctypes.c_char_p,
                                                      ctypes.c_int,
                                                      POINTER(TessResultRenderer)]
        tesseract.TessBaseAPIProcessPages.restype = ctypes.c_int

        tesseract.TessBaseAPIAnalyseLayout.argtypes = [POINTER(TessBaseAPI)]
        tesseract.TessBaseAPIAnalyseLayout.restype = POINTER(TessPageIterator)

        tesseract.TessPageIteratorIsAtBeginningOf.argtypes = [POINTER(TessPageIterator),
                                                              ctypes.c_int]
        tesseract.TessPageIteratorIsAtBeginningOf.restype = ctypes.c_int

        tesseract.TessPageIteratorBoundingBox.argtypes = [POINTER(TessPageIterator),
                                                          ctypes.c_int,
                                                          POINTER(ctypes.c_int),
                                                          POINTER(ctypes.c_int),
                                                          POINTER(ctypes.c_int),
                                                          POINTER(ctypes.c_int)]
        tesseract.TessPageIteratorBoundingBox.restype = ctypes.c_int

        tesseract.TessBaseAPIGetIterator.argtypes = [POINTER(TessBaseAPI)]
        tesseract.TessBaseAPIGetIterator.restype = POINTER(TessResultIterator)

        tesseract.TessResultIteratorGetPageIterator.argtypes = [POINTER(TessResultIterator)]
        tesseract.TessResultIteratorGetPageIterator.restype = POINTER(TessPageIterator)

        tesseract.TessResultIteratorNext.argtypes = [POINTER(TessResultIterator), ctypes.c_int]
        tesseract.TessResultIteratorNext.restype = ctypes.c_int

        tesseract.TessResultIteratorDelete.argtypes = [POINTER(TessResultIterator)]
        tesseract.TessResultIteratorDelete.restype = None

        tesseract.TessPageIteratorDelete.argtypes = [POINTER(TessPageIterator)]
        tesseract.TessPageIteratorDelete.restype = None

        leptonica.pixRead.argtypes = [ctypes.c_char_p]
        leptonica.pixRead.restype = POINTER(Pix)

        leptonica.pixDestroy.argtypes = [POINTER(POINTER(Pix))]
        leptonica.pixDestroy.restype = None

        tesseract.TessBaseAPIGetHOCRText.argtypes = [POINTER(TessBaseAPI), ctypes.c_int]
        tesseract.TessBaseAPIGetHOCRText.restype = ctypes.c_char_p

        tesseract.TessBaseAPIGetAvailableLanguagesAsVector.argtypes = [POINTER(TessBaseAPI)]
        tesseract.TessBaseAPIGetAvailableLanguagesAsVector.restype = POINTER(ctypes.c_char_p)

        # fill in available tesseract classifiers as they are only determinable
        # after setting the tessdata directory.
        ocr_tesseract.arg_values['languages'] = _get_available_classifiers()

@app.task(base=NidabaTask, name=u'nidaba.segmentation.tesseract')
def segmentation_tesseract(doc, method=u'segment_tesseract'):
    """
    Performs page segmentation using tesseract's built-in algorithm and writes
    a TEI XML segmentation file.

    Args:
        doc (unicode, unicode): The input document tuple
        method (unicode): The suffix string appended to all output files.

    Returns:
        Two storage tuples with the first one containing the segmentation and
        the second one being the file the segmentation was calculated upon.
    """
    input_path = storage.get_abs_path(*doc)
    output_path = splitext(storage.insert_suffix(input_path, method))[0] + '.xml'

    ver = tesseract.TessVersion()
    if int(ver.split('.')[0]) < 3 or int(ver.split('.')[1]) < 2:
        raise NidabaTesseractException('libtesseract version is too old. Set '
                                       'implementation to direct.')
    api = tesseract.TessBaseAPICreate()
    rc = tesseract.TessBaseAPIInit3(api, tessdata.encode('utf-8'), None)
    if (rc):
        tesseract.TessBaseAPIDelete(api)
        raise NidabaTesseractException('Tesseract initialization failed.')

    # only do segmentation and script detection
    logger.debug('Setting page set mode to 2')
    tesseract.TessBaseAPISetPageSegMode(api, 2)

    logger.debug('Reading {} using leptonica'.format(input_path))
    pix = leptonica.pixRead(input_path.encode('utf-8'))
    logger.debug('Setting PIX as input image')
    tesseract.TessBaseAPISetImage2(api, pix)
    logger.debug('Analyzing page layout')
    it = tesseract.TessBaseAPIAnalyseLayout(api)
    logger.debug('Destroying PIX')
    leptonica.pixDestroy(ctypes.byref(pix))
    x0, y0, x1, y1 = (ctypes.c_int(), ctypes.c_int(), ctypes.c_int(),
                      ctypes.c_int())

    w, h = Image.open(input_path).size
    logger.info('Initializing TEI XML file with {}x{} {}/{}'.format(w, h, *doc))
    tei = TEIFacsimile()
    tei.document((w, h), os.path.join(*doc))
    tei.title = os.path.basename(doc[1])
    tei.add_respstmt('tesseract', 'page segmentation')

    while True:
        if tesseract.TessPageIteratorIsAtBeginningOf(it, RIL_TEXTLINE):
            tesseract.TessPageIteratorBoundingBox(it,
                                                  RIL_TEXTLINE,
                                                  ctypes.byref(x0),
                                                  ctypes.byref(y0),
                                                  ctypes.byref(x1),
                                                  ctypes.byref(y1))
            tei.add_line((x0.value, y0.value, x1.value, y1.value))
            logger.debug('Segmenter found new line at {} {} {} {}'.format(x0.value,
                                                                          y0.value,
                                                                          x1.value,
                                                                          y1.value))
        if tesseract.TessPageIteratorIsAtBeginningOf(it, RIL_WORD):
            tesseract.TessPageIteratorBoundingBox(it,
                                                  RIL_WORD,
                                                  ctypes.byref(x0),
                                                  ctypes.byref(y0),
                                                  ctypes.byref(x1),
                                                  ctypes.byref(y1))
            tei.add_segment((x0.value, y0.value, x1.value, y1.value))
            logger.debug('Segmenter found new word at {} {} {} {}'.format(x0.value,
                                                                          y0.value,
                                                                          x1.value,
                                                                          y1.value))

        tesseract.TessPageIteratorBoundingBox(it,
                                              RIL_SYMBOL,
                                              ctypes.byref(x0),
                                              ctypes.byref(y0),
                                              ctypes.byref(x1),
                                              ctypes.byref(y1))
        tei.add_graphemes([(None, (x0.value, y0.value, x1.value, y1.value))])
        logger.debug('Segmenter found new symbol at {} {} {} {}'.format(x0.value,
                                                                        y0.value,
                                                                        x1.value,
                                                                        y1.value))
        if not tesseract.TessPageIteratorNext(it, RIL_SYMBOL):
            logger.debug('No more elements on page')
            break
    logger.debug('Deleting page iterator and base API')
    tesseract.TessPageIteratorDelete(it)
    tesseract.TessBaseAPIEnd(api)
    tesseract.TessBaseAPIDelete(api)
    logger.info('Writing segmentation to {}'.format(output_path))
    with open(output_path, 'w') as fp:
        tei.write(fp)
    return storage.get_storage_path(output_path), doc


def _get_available_classifiers():
    api = tesseract.TessBaseAPICreate()
    tesseract.TessBaseAPIInit3(api, tessdata.encode('utf-8'), None)
    classifiers = tesseract.TessBaseAPIGetAvailableLanguagesAsVector(api)
    ret = []
    for cls in classifiers:
        if cls is None:
            break
        ret.append(cls)
    tesseract.TessDeleteTextArray(classifiers)
    return ret


@app.task(base=NidabaTask, name=u'nidaba.ocr.tesseract',
          arg_values={'languages': None, # is filled by the setup function
                      'extended': [False, True]})
def ocr_tesseract(doc, method=u'ocr_tesseract', languages=None,
                  extended=False):
    """
    Runs tesseract on an input document.

    Args:
        doc (unicode, unicode): The input document tuple
        method (unicode): The suffix string appended to all output files
        languages (list): A list of tesseract classifier identifiers
        extended (bool): Switch to enable extended hOCR generation containing
                         character cuts and confidences. Has no effect when
                         direct or legacy implementation is used.

    Returns:
        (unicode, unicode): Storage tuple for the output file
    """
    image_path = storage.get_abs_path(*doc[1])

    # rewrite the segmentation file to lines in UZN format
    logger.debug('Rewriting TEI ({}) -> UZN ({})'.format(doc[0][1],
                                                         splitext(doc[1][1])[0]
                                                         + '.uzn'))
    seg = TEIFacsimile()
    with storage.StorageFile(*doc[0]) as fp:
        seg.read(fp)
    with storage.StorageFile(doc[1][0], splitext(doc[1][1])[0] + '.uzn', mode='wb') as fp:
        uzn = UZNWriter(fp)
        for line in seg.lines:
            uzn.writerow(*line[:4])

    if isinstance(languages, basestring):
        languages = [languages]
    output_path = storage.insert_suffix(image_path, method, *languages)

    logger.debug('Invoking tesseract with {} call method'.format(implementation))
    if implementation == 'legacy':
        result_path = output_path + '.html'
        ocr_direct(image_path, output_path, languages)
    elif implementation == 'direct':
        result_path = output_path + '.hocr'
        ocr_direct(image_path, output_path, languages)
    elif implementation == 'capi':
        result_path = output_path + '.xml'
        ocr_capi(image_path, result_path, seg, languages, extended)
    else:
        raise NidabaTesseractException('Invalid implementation selected',
                                       implementation)

    if not result_path[-4:] == '.xml':
        logger.debug('Converting hOCR ({}) -> TEI ({})'.format(result_path,
                                                               output_path +
                                                               '.xml'))
        tei = TEIFacsimile()
        with open(result_path) as fp:
            tei.load_hocr(fp)
        os.unlink(result_path)
        with open(output_path + '.xml', 'wb') as fp:
            tei.write(fp)
        result_path = output_path + '.xml'
    return storage.get_storage_path(result_path)


def ocr_capi(image_path, output_path, facsimile, languages, extended=False):
    """
    OCRs an image using the C API provided by tesseract versions 3.02 and
    higher.

    Args:
        image_path (unicode): Path to the input image
        facsimile (nidaba.tei.TEIFacsimile): Facsimile object of the
                                             segmentation
        output_path (unicode): Path to the hOCR output
        languages (list): List of valid tesseract language identifiers
        extended (bool): Switch to select extended hOCR output containing
                         character cuts and confidences values
    """
    # tesseract has a tendency to crash arbitrarily on some inputs
    # necessitating execution in a separate process to ensure the worker
    # doesn't just die. We use fork as the multiprocessing module thinks
    # programmers are too stupid to reap their children.
    logger.info('Forking before entering unstable ctypes code')
    pid = os.fork()
    if pid != 0:
        try:
            logger.info('Waiting for child to complete')
            _, status = os.waitpid(pid, 0)
        except OSError as e:
            if e.errno not in (errno.EINTR, errno.ECHILD):
                raise
            return
        if os.WIFSIGNALED(status):
            raise NidabaTesseractException('Tesseract killed by signal: {0}'.format(os.WTERMSIG(status)))
        return

    # ensure we've loaded a tesseract object newer than 3.02
    ver = tesseract.TessVersion()
    if int(ver.split('.')[0]) < 3 or int(ver.split('.')[1]) < 2:
        raise NidabaTesseractException('libtesseract version is too old. Set '
                                       'implementation to direct.')

    logger.info('Creating base API (tessdata: {}, classifiers: {})'.format(tessdata,
                                                                           '+'.join(languages)))
    api = tesseract.TessBaseAPICreate()
    rc = tesseract.TessBaseAPIInit3(api, tessdata.encode('utf-8'),
                                    ('+'.join(languages)).encode('utf-8'))
    if (rc):
        tesseract.TessBaseAPIDelete(api)
        raise NidabaTesseractException('Tesseract initialization failed.')

    logger.debug('Reading {} using leptonica'.format(image_path))
    pix = leptonica.pixRead(image_path.encode('utf-8'))
    logger.debug('Set PIX as source image.')
    tesseract.TessBaseAPISetImage2(api, pix)
    if extended:
        logger.debug('Set page segmentation mode to single line (extended output)')
        tesseract.TessBaseAPISetPageSegMode(api, 7)
        facsimile.add_respstmt('tesseract', 'character recognition')

        # While tesseract can recognize single words/characters it is extremely
        # slow to do so. We therefore wrote an UZN file containing only lines
        # and clear out segments and graphemes here too.
        logger.debug('Clearing unused elements from segmentation')
        facsimile.clear_graphemes()
        facsimile.clear_segments()

        w, h = Image.open(image_path).size
        x0, y0, x1, y1 = (ctypes.c_int(), ctypes.c_int(), ctypes.c_int(),
                          ctypes.c_int())
        for line in facsimile.lines:
            logger.debug('Clipping input image to recognition box ({})'.format(line[4]))
            tesseract.TessBaseAPISetRectangle(api, line[0], line[1], line[2] - line[0], line[3] - line[1])
            logger.debug('Recognizing line.')
            if tesseract.TessBaseAPIRecognize(api, None):
                leptonica.pixDestroy(ctypes.byref(pix))
                tesseract.TessBaseAPIEnd(api)
                tesseract.TessBaseAPIDelete(api)
                raise NidabaTesseractException('Recognition failed.')
            logger.debug('Retrieving result and page iterator.')
            ri = tesseract.TessBaseAPIGetIterator(api)
            pi = tesseract.TessResultIteratorGetPageIterator(ri)
            logger.debug('Scoping line {}'.format(line[4]))
            facsimile.scope_line(line[4])
            last_word = None
            while True:
                if tesseract.TessPageIteratorIsAtBeginningOf(pi, RIL_WORD):
                    lang = tesseract.TessResultIteratorWordRecognitionLanguage(ri, RIL_WORD).decode('utf-8')
                    tesseract.TessPageIteratorBoundingBox(pi, RIL_WORD,
                                                          ctypes.byref(x0),
                                                          ctypes.byref(y0),
                                                          ctypes.byref(x1),
                                                          ctypes.byref(y1))
                    conf = tesseract.TessResultIteratorConfidence(ri, RIL_WORD)
                    # insert space between word boundaries as they don't get
                    # returned by GetUTF8Text
                    if last_word is not None:
                        facsimile.clear_segment()
                        facsimile.add_graphemes([(u' ', (last_word,
                                                         y0.value,
                                                         x0.value,
                                                         y1.value), 100.0)])
                    last_word = x1.value
                    facsimile.add_segment((x0.value, y0.value, x1.value, y1.value),
                                          lang, conf)
                    logger.debug('New word at {} {} {} {} (conf: {})'.format(x0.value,
                                                                             y0.value,
                                                                             x1.value,
                                                                             y1.value,
                                                                             conf))

                conf = tesseract.TessResultIteratorConfidence(ri, RIL_SYMBOL)
                tesseract.TessPageIteratorBoundingBox(pi, RIL_SYMBOL,
                                                      ctypes.byref(x0),
                                                      ctypes.byref(y0),
                                                      ctypes.byref(x1),
                                                      ctypes.byref(y1))
                grapheme = tesseract.TessResultIteratorGetUTF8Text(ri, RIL_SYMBOL)
                if grapheme is not None:
                    grapheme = grapheme.decode('utf-8')
                facsimile.add_graphemes([(grapheme, (x0.value, y0.value, x1.value,
                                          y1.value), conf)])
                logger.debug('New symbol {} at {} {} {} {} (conf: {})'.format(grapheme,
                                                                              x0.value,
                                                                              y0.value,
                                                                              x1.value,
                                                                              y1.value,
                                                                              conf))
                if not tesseract.TessResultIteratorNext(ri, RIL_SYMBOL):
                    logger.debug('No more symbols on page')
                    break
            # XXX: deleting the page iterator too hangs
            logger.debug('Deleting result iterator')
            tesseract.TessResultIteratorDelete(ri)
        logger.info('Writing TEI ({})'.format(output_path))
        with open(output_path, 'wb') as fp:
            facsimile.write(fp)
    else:
        logger.debug('Set page segmentation to single column')
        tesseract.TessBaseAPISetPageSegMode(api, 4)
        logger.debug('Recognizing page')
        if tesseract.TessBaseAPIRecognize(api, None):
            leptonica.pixDestroy(ctypes.byref(pix))
            tesseract.TessBaseAPIDelete(api)
            raise NidabaTesseractException('Tesseract recognition failed')
        with open(output_path, 'wb') as fp:
            logger.debug('Retrieving hOCR')
            tp = tesseract.TessBaseAPIGetHOCRText(api, 0)
            hocr = StringIO.StringIO()
            hocr.write(ctypes.string_at(tp))
            hocr.seek(0)
            logger.debug('Converting hOCR -> TEI ({})'.format(output_path))
            facsimile.load_hocr(hocr)
            facsimile.write(fp)
            logger.debug('Freeing hOCR string')
            tesseract.TessDeleteText(tp)
    logger.debug('Deleting base API')
    tesseract.TessBaseAPIEnd(api)
    tesseract.TessBaseAPIDelete(api)
    logger.info('Quitting child process')
    os._exit(os.EX_OK)


def ocr_direct(image_path, output_path, languages):
    """
    OCRs an image by calling the tesseract executable directly. Images are read
    using the linked leptonica library and the given output_path WILL be
    modified by tesseract.

    Args:
        image_path (unicode): Path to the input image
        output_path (unicode): Path to the hOCR output
        languages (list): List of valid tesseract language identifiers
    """
    logger.debug('Calling tesseract executable in subprocess')
    p = subprocess.Popen(['tesseract', image_path, output_path, '-l',
                          '+'.join(languages), '-psm', '4', '--tessdata-dir',
                          tessdata, 'hocr'], stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE)
    out, err = p.communicate()
    if p.returncode:
        raise NidabaTesseractException(err)



# -*- coding: utf-8 -*-
"""
nidaba.nidaba
~~~~~~~~~~~~~

The public API of nidaba. External applications should exclusively use the
objects and methods defined here.
"""

from __future__ import unicode_literals, print_function, absolute_import

from nidaba.nidabaexceptions import (NidabaInputException,
                                     NidabaNoSuchAlgorithmException,
                                     NidabaTickException, NidabaStepException)

from celery import chain
from celery import group
from itertools import product
from inspect import getcallargs
from collections import OrderedDict, Iterable
from requests_toolbelt.multipart import encoder

import os
import json
import uuid
import requests


def task_arg_validator(arg_values, **kwargs):
    """
    Validates keyword arguments against the list of valid argument values
    contained in the task definition.

    Raises:
        NidabaInputException if validation failed.
    """
    kwc = kwargs.copy()

    def _val_single_arg(arg, type):
        if type == 'float':
            if not isinstance(val, float) and not isinstance(val, int):
                raise NidabaInputException('{} is not a float'.format(val))
        elif type == 'int':
            if not isinstance(val, int):
                raise NidabaInputException('{} is not an int'.format(val))
        elif type == 'str':
            if not isinstance(val, basestring):
                raise NidabaInputException('{} is not a string'.format(val))
        # XXX: Add file checker for local case
        elif type == 'file':
            pass
        else:
            raise NidabaInputException('Argument type {} unknown'.format(type))

    for k, v in arg_values.iteritems():
        try:
            val = kwc.pop(k)
        except:
            raise NidabaInputException('Missing argument: {}'.format(k))
        if isinstance(v, tuple):
            if not isinstance(val, type(v[0])):
                raise NidabaInputException('{} of different type than range fields'.format(val))
            if val < v[0] or val > v[1]:
                raise NidabaInputException('{} outside of allowed range {}-{}'.format(val, *v))
        elif isinstance(v, list):
            if isinstance(val, Iterable) and not isinstance(val, basestring):
                va = set(val)
            else:
                va = set([val])
            if not set(v).issuperset(va):
                raise NidabaInputException('{} not in list of valid values'.format(val))
        else:
            _val_single_arg(val, v)

    if kwc:
        raise NidabaInputException('Superfluous arguments present')


class Batch(object):

    """
    Creates a series of celery tasks OCRing a set of documents (among other
    things).

    A batch contains three level of definitions: tasks, ticks, and steps. A
    task is a singular operation on an input document, creating a single output
    document, e.g. binarization using a particular configuration of an
    algorithm or OCR using a particular engine. Multiple tasks executing in
    parallel are grouped into a tick and multiple ticks (running sequentially)
    are grouped into steps which again are executed sequentially.

    The need for steps and ticks arises from two different execution orders
    required by a pipeline. Take the following example:

        step 1:
            tick a: task 1
            tick b: task 2, task 3
            tick c: task 4, task 5

    The pipeline expands this example to the following sequences run in
    parallel (dot product of all ticks):

        task 1 -> task 2 -> task 4
        task 1 -> task 2 -> task 5
        task 1 -> task 3 -> task 4
        task 1 -> task 3 -> task 5

    It is not garantueed that any particular task in another sequence has
    executed successfully before a task in a sequence is run, i.e. it is not
    ensured that all task 1's have finished before task 2 of the first sequence
    is executed, except the task(s) further up the sequence.

    Steps on the other hand ensure that all tasks of the previous step have
    finished successfully. The output(s) of the expanded ticks is aggregated
    into a single list and used as the input of the first tick of the step.
    Expanding on the example the following step is added:

        step 2:
            tick d: task 6

    After the 4 sequence are finished their output is aggregated into a list
    [d1, d2, d3, d4] and used as the input of task 6. The final output of task
    6 is the output of the pipeline.

    The call order to create this example is:

        Batch.add_step()
        Batch.add_tick()
        Batch.add_task(task_1)
        Batch.add_tick()
        Batch.add_task(task_2)
        Batch.add_task(task_3)
        Batch.add_tick()
        Batch.add_task(task_4)
        Batch.add_task(task_5)
        Batch.add_step()
        Batch.add_tick()
        Batch.add_task(task_6)
    """

    def __init__(self, id):
        # stuff depending on a valid configuration
        from nidaba import storage
        from nidaba import config
        self.storage = storage

        # slowly importing stuff
        from nidaba import tasks
        from nidaba import plugins
        from nidaba import celery
        self.tasks = tasks
        self.celery = celery

        self.id = id
        self.docs = []
        self.batch_def = []
        self.cur_step = None
        self.cur_tick = None
        self.scratchpad = {}
        self.redis = config.Redis


    def add_scratchpad(self):
        """
        Adds a scratchpad to the database that allows suspending and resuming
        task assembly. 
        
        All added documents and tasks will be stored persistently;
        reinstantiating the Batch object and calling restore_scratchpad will
        restore the object to the previous state.

        The scratchpad will be destroyed on task execution.

        Raises:
            NidabaInputException if a scratchpad is already attached or the
            batch has been executed.
        """
        if self.redis.get(self.id) is not None:
            raise NidabaInputException('Batch already has scratchpad or been '
                                       'executed.')
        else:
            self.scratchpad = {'scratchpad': {'docs': self.docs, 
                                              'batch_def': self.batch_def,
                                              'cur_step': self.cur_step,
                                              'cur_tick': self.cur_tick}}
            self.redis.set(self.id, json.dumps(self.scratchpad))

    def restore_scratchpad(self):
        """
        Retrieves the scratchpad from the database and restores its contents
        to the current Batch object overwriting any information defined in it.

        Raises:
            NidabaInputException if no scratchpad can be found either because
            none has been attached or the batch has already been executed.
        """
        scratch = json.loads(self.redis.get(self.id))
        if scratch and 'scratchpad' in scratch:
            self.scratchpad = scratch
            for k, v in self.scratchpad['scratchpad'].iteritems():
                setattr(self, k, v)
        else:
            raise NidabaInputException('No scratchpad in database')

    def get_state(self):
        """
        Retrieves the current state of a batch.

        Returns:
            (unicode): A string containing one of the following states:

                NONE: The batch ID is not registered in the backend.
                FAILURE: Batch execution has failed.
                PENDING: The batch is currently running.
                SUCCESS: The batch has completed successfully.
        """
        batch = self.redis.get(self.id)
        try:
            batch = json.loads(batch)
        except Exception:
            return u'NONE'
        if 'scratchpad' in batch:
            return u'NONE'
        st = 'SUCCESS'
        for subtask in batch.itervalues():
            if subtask['state'] == 'PENDING' or subtask['state'] == 'RUNNING':
                st = u'PENDING'
            if subtask['state'] == 'FAILURE':
                return u'FAILURE'
        return st

    def get_errors(self):
        """
        Retrieves all errors of the batch.

        Returns:
            list: A list of tuples containing keyword arguments to the task, a
            dictionary containing debug tracking information (i.e. variables
            which are given to the tasks as keyword arguments but aren't
            arguments to the underlying function), and the exception message of
            the failure.
        """
        batch = self.redis.get(self.id)
        try:
            batch = json.loads(batch)
        except:
            return None
        errors = []
        for subtask in batch.itervalues():
            if subtask['state'] == 'FAILURE':
                errors.append(subtask)
        return errors

    def get_results(self):
        """
        Retrieves the storage tuples of a successful batch.

        Returns:
        """
        batch = self.redis.get(self.id)
        try:
            batch = json.loads(batch)
        except Exception:
            return None

        outfiles = []
        for subtask in batch.itervalues():
            if len(subtask['children']) == 0 and not subtask['housekeeping'] and subtask['result'] is not None:
                outfiles.append((subtask['result'], subtask['root_document']))
        return outfiles

    def get_extended_state(self):
        """
        Returns extended batch state information.

        Returns:
            A dictionary containing an entry for each subtask.
        """
        state = json.loads(self.redis.get(self.id))
        if 'scratchpad' in state:
            return []
        else:
            return state

    def add_document(self, doc):
        """Add a document to the batch.

        Adds a document tuple to the batch and checks if it exists.

        Args:
            doc (tuple): A standard document tuple.

        Raises:
            NidabaInputException: The document tuple does not refer to a file.
        """

        if not self.storage.is_file(*doc):
            raise NidabaInputException('Input document is not a file.')
        self.docs.append(doc)
        if self.scratchpad:
            self.scratchpad['scratchpad']['docs'] = self.docs
            self.redis.set(self.id, json.dumps(self.scratchpad))

    def add_task(self, method, **kwargs):
        """Add a task to the current tick.

        Adds a ``task``, a single executable task gathering one or more input
        documents and returning a single output document, to the current tick.
        Multiple jobs are run in parallel.

        Args:
            method (unicode): A task identifier
            **kwargs: Arguments to the task

        Raises:
            NidabaTickException: There is no tick to add a task to.
            NidabaNoSuchAlgorithmException: Invalid method given.
        """
        if self.cur_tick is None:
            raise NidabaTickException('No tick to add task to.')
        if u'nidaba.' + method not in self.celery.app.tasks:
            raise NidabaNoSuchAlgorithmException('No such task in registry')
        kwargs[u'method'] = method
        self.cur_tick.append(kwargs)
        if self.scratchpad:
            self.scratchpad['scratchpad']['cur_tick'].append(kwargs)
            self.redis.set(self.id, json.dumps(self.scratchpad))


    def add_tick(self):
        """Add a new tick to the current step.

        Adds a ``tick``, a set of tasks running in parallel and sharing common
        input documents to the current step.

        Raises:
            NidabaStepException: There is no step to add a tick to.
        """
        if self.cur_step is None:
            raise NidabaStepException('No step to add tick to.')
        if self.cur_tick:
            self.cur_step.append(self.cur_tick)
        self.cur_tick = []
        if self.scratchpad:
            self.scratchpad['scratchpad']['cur_step'] = self.cur_step
            self.scratchpad['scratchpad']['cur_tick'] = []
            self.redis.set(self.id, json.dumps(self.scratchpad))


    def add_step(self):
        """Add a new step to the batch definition.

        Adds a ``step``, a list of sequentially run ticks to the batch
        definition. The output(s) of the last tick of a step is aggregated into
        a single list and used as the input of the first tick of the following
        step.
        """
        if self.cur_tick:
            self.cur_step.append(self.cur_tick)
            self.cur_tick = []
        if self.cur_step:
            self.batch_def.append(self.cur_step)
        self.cur_step = []
        if self.scratchpad:
            self.scratchpad['scratchpad']['cur_step'] = self.cur_step
            self.scratchpad['scratchpad']['cur_tick'] = self.cur_tick
            self.scratchpad['scratchpad']['batch_def'] = self.batch_def
            self.redis.set(self.id, json.dumps(self.scratchpad))


    def run(self):
        """Executes the current batch definition.

        Expands the current batch definition to a series of celery chains and
        executes them asynchronously. Additionally a batch record is written to
        the celery result backend.

        Returns:
            (unicode): Batch identifier.
        """
        # flush most recent tick/step onto the batch definition
        self.add_tick()
        self.add_step()

        groups = []
        tick = []

        result_data = {}

        # We first expand the tasks starting from the second step as these are
        # the same for each input document.
        if len(self.batch_def) > 1:
            groups.append(self.tasks.util.sync.s())
            for tset in self.batch_def[1:]:
                for sequence in product(*tset):
                    method = self.celery.app.tasks['nidaba.' +
                                                   sequence[0]['method']]
                    ch = chain(method.s(**(sequence[0])))
                    for seq in sequence[1:]:
                        method = self.celery.app.tasks['nidaba.' + seq['method']]
                        ch |= method.s(**seq)
                    tick.append(ch)
                groups.append(group(tick))
                groups.append(self.tasks.util.sync.s())

        # The expansion steps described above is redone for each input document
        chains = []
        for doc in self.docs:
            tick = []
            # the first step is handled differently as the input document has
            # to be added explicitely.
            for sequence in product([doc], *self.batch_def[0]):
                method = self.celery.app.tasks['nidaba.' + sequence[1]['method']]
                root = sequence[0]
                ch = chain(method.s(doc=root, **(sequence[1])))
                for seq in sequence[2:]:
                    method = self.celery.app.tasks['nidaba.' + seq['method']]
                    ch |= method.s(**seq)
                tick.append(ch)
            doc_group = group(tick)
            parents = []
            group_list = [doc_group] + groups
            chains.append(chain(group_list))

            # Now we give out another set of unique task identifiers and save
            # them to the database. Presetting the celery task ID does not work
            # as our chains get automagically upgraded to chords by celery,
            # erasing the ID fields in the process.
            for step in enumerate(group_list):
                if hasattr(step[1], 'tasks'):
                    for tick in enumerate(step[1].tasks):
                        if step[0]:
                            parents = parent_step
                        else:
                            parents = []
                        for task in enumerate(tick[1].tasks):
                            task_id = uuid.uuid4().get_hex()
                            result_data[task_id] = {
                                'children': [],
                                'parents': parents,
                                'housekeeping': False,
                                'root_document': doc,
                                'state': 'PENDING',
                                'result': None,
                                'task': (task[1]['task'], task[1]['kwargs']),
                            }
                            for parent in parents:
                                result_data[parent]['children'] = task_id
                            group_list[step[0]].tasks[tick[0]].tasks[task[0]].kwargs['task_id'] = task_id
                            group_list[step[0]].tasks[tick[0]].tasks[task[0]].kwargs['batch_id'] = self.id
                            parents = [task_id]
                    parent_step = [x.tasks[-1].kwargs['task_id'] for x in
                                   step[1].tasks]
                else:
                    task_id = uuid.uuid4().get_hex()
                    result_data[task_id] = {
                        'children': [],
                        'parents': [],
                        'housekeeping': True,
                        'root_document': doc,
                        'state': 'PENDING',
                        'result': None,
                        'task': (task[1]['task'], task[1]['kwargs']),
                    }
                    group_list[step[0]].kwargs['task_id'] = task_id
                    group_list[step[0]].kwargs['batch_id'] = self.id
        # also deletes the scratchpad 
        self.redis.set(self.id, json.dumps(result_data))
        [x.apply_async() for x in chains]
        return self.id


class SimpleBatch(Batch):
    """
    A simpler interface to the batch functionality that is more amenable to
    RESTful task assembly and prevents some incidences of
    bullet-in-foot-syndrome.

    A SimpleBatch contains only a list of input documents and a series of
    tasks. The order of task execution depends on a predefined order, similar
    to the ``nidaba`` command-line util. 

    If no batch identifier is given a new batch will be created.

    SimpleBatches always contain a scratchpad (which will be restored
    automatically).
    """
    def __init__(self, id=None):
        # stuff depending on a valid configuration
        from nidaba import storage
        self.storage = storage

        # slowly importing stuff (tasks and plugins also needed so the task
        # registry is complete)
        from nidaba import tasks
        from nidaba import plugins
        from nidaba import celery
        self.celery = celery

        if id is None:
            id = unicode(uuid.uuid4())
            self.storage.prepare_filestore(id)
        if not self.storage.is_valid_job(id):
            raise NidabaInputException('Storage not prepared for task')
        super(SimpleBatch, self).__init__(id)
        self.lock = False
        
        keys = ['img', 'binarize', 'segmentation', 'ocr', 'stats', 'postprocessing', 'output']
        self.tasks = OrderedDict([(key, []) for key in keys])
        try:
            self.add_scratchpad()
            self.scratchpad['scratchpad']['tasks'] = self.tasks
            self.redis.set(self.id, json.dumps(self.scratchpad))
        except:
            try:
                self.restore_scratchpad()
                # reorder the tasks dictionary
                self.tasks = OrderedDict([(key, self.tasks[key]) for key in  keys])
            except NidabaInputException:
                # no scratchpad in database means batch is running and may not
                # be modified.
                self.lock = True

    def is_running(self):
        """
        Returns True if the batch's run() method has been successfully called, otherwise False.
        """
        return self.lock

    def get_tasks(self):
        """
        Returns the simplified task definition either from the scratchpad or
        from the pipeline when already in execution.
        """
        entry = json.loads(self.redis.get(self.id))
        if 'scratchpad' in entry:
            scratch = json.loads(self.redis.get(self.id))
            return scratch['scratchpad']['tasks']
        else:
            state = super(SimpleBatch, self).get_extended_state()
            tasks = OrderedDict([('img', []), 
                                 ('binarize', []),
                                 ('segmentation', []), 
                                 ('ocr', []),
                                 ('stats', []), 
                                 ('postprocessing', []),
                                 ('output', [])])
            for task in state.itervalues():
                _, group, method = task['task'][0].split('.')
                if group in tasks:
                    tasks[group].append(('{}.{}'.format(group, method),
                                         task['task'][1]))
            return tasks

    def get_documents(self):
        """
        Returns the list of input document for this task.
        """
        entry = json.loads(self.redis.get(self.id))
        if 'scratchpad' in entry:
            return entry['scratchpad']['docs']
        else:
            state = super(SimpleBatch, self).get_extended_state()
            docs = []
            for task in state.itervalues():
                if task['root_document'] not in docs:
                    docs.append(task['root_document'])
            return docs

    @staticmethod
    def get_available_tasks():
        """
        Returns all available tasks and their valid argument values.

        The return value is an ordered dictionary containing an entry for each
        group with a sub-dictionary containing the task identifiers and valid
        argument values.
        """
        tasks = OrderedDict()
        from nidaba import celery
        for task, fun in celery.app.tasks.iteritems():
            try:
                _, group, method = task.split('.')
            except:
                continue
            if group not in tasks:
                tasks[group] = {}
            kwargs = fun.get_valid_args()
            tasks[group][method] = kwargs
        return tasks

    def add_document(self, doc):
        """Add a document to the batch.

        Adds a document tuple to the batch and checks if it exists.

        Args:
            doc (tuple): A standard document tuple.

        Raises:
            NidabaInputException: The document tuple does not refer to a file
                                  or the batch is locked because the run()
                                  method has been called.
        """
        if self.lock:
            raise NidabaInputException('Executed batch may not be modified')
        super(SimpleBatch, self).add_document(doc)


    def add_task(self, group, method, **kwargs):
        """
        Add a particular task configuration to a task group.

        Args:
            group (unicode): Group the task belongs to
            method (unicode): Name of the task
            kwargs: Arguments to the task
        """
        if self.lock:
            raise NidabaInputException('Executed batch may not be modified')
        # validate that the task exists
        if group not in self.tasks:
            raise NidabaNoSuchAlgorithmException('Unknown task group')
        if u'nidaba.{}.{}'.format(group, method) not in self.celery.app.tasks:
            raise NidabaNoSuchAlgorithmException('Unknown task')
        task = self.celery.app.tasks[u'nidaba.{}.{}'.format(group, method)]
        # validate arguments first against getcallargs
        try:
            getcallargs(task.run, ('', ''), **kwargs)
        except TypeError as e:
            raise NidabaInputException(str(e))
        # validate against arg_values field of the task
        task_arg_validator(task.get_valid_args(), **kwargs)
        self.tasks[group].append((method, kwargs))
        self.scratchpad['scratchpad']['tasks'] = self.tasks
        self.redis.set(self.id, json.dumps(self.scratchpad))


    def run(self):
        """Executes the current batch definition.

        Expands the current batch definition to a series of celery chains and
        executes them asynchronously. Additionally a batch record is written to
        the celery result backend.

        Returns:
            (unicode): Batch identifier.
        """
        if self.lock:
            raise NidabaInputException('Executed batch may not be modified')

        self.add_step()
        for group, btasks in self.tasks.iteritems():
            self.add_tick()
            for task in btasks:
                super(SimpleBatch, self).add_task('{}.{}'.format(group,
                                                                 task[0]),
                                                  **task[1])
        self.lock = True
        return super(SimpleBatch, self).run()


class NetworkSimpleBatch(object):
    """
    A SimpleBatch object providing the same interface as a SimpleBatch.

    It does some basic error checking to minimize network traffic but it won't
    catch all errors before issuing API requests, especially if the batch is
    modified by another process. In these cases exceptions will be raised by
    the ``requests`` module.
    """
    def __init__(self, host, id=None):
        self.id = id
        self.host = host
        self.lock = False
        self.allowed_tasks = {}
        if id is not None:
            r = requests.get('{}/batch/{}'.format(host, id))
            r.raise_for_status()


    def create_batch(self):
        """
        Creates a batch on the server. Also synchronizes the list of available
        tasks and their parameters.
        """
        if self.id is not None:
            raise NidabaInputException('SimpleBatch object already initialized')
        r = requests.post('{}/batch'.format(self.host))
        r.raise_for_status()
        self.id = r.json()['id']
        self.lock = False
        self.get_available_tasks()
        return self.id

    def get_available_tasks(self):
        """
        Synchronizes the local task/parameter list with the remote server.
        """
        r = requests.get('{}/tasks'.format(self.host))
        r.raise_for_status()
        self.allowed_tasks = r.json()

    def is_running(self):
        """
        Returns True if the batch's run() method has been successfully called, otherwise False.
        """
        if not self.id:
            raise NidabaInputException('Object not attached to batch.')
        r = requests.get('{}/batch/{}'.format(self.host, self.id))
        r.raise_for_status()
        self.lock = True
        if r.json():
            self.lock = True
            return True
        else:
            self.lock = False
            return False

    def get_state(self):
        """
        Retrieves the current state of a batch.

        Returns:
            (unicode): A string containing one of the following states:

                NONE: The batch ID is not registered in the backend.
                FAILURE: Batch execution has failed.
                PENDING: The batch is currently running.
                SUCCESS: The batch has completed successfully.
        """
        if not self.id:
            raise NidabaInputException('Object not attached to batch.')
        r = requests.get('{}/batch/{}'.format(self.host, self.id))
        r.raise_for_status()
        batch = r.json()
        if 'scratchpad' in batch:
            return u'NONE'
        elif 'chains' in batch:
            self.lock = True
            batch = batch['chains']
            st = u'SUCCESS'
            for subtask in batch.itervalues():
                if subtask['state'] == 'PENDING' or subtask['state'] == 'RUNNING':
                    st = u'PENDING'
                if subtask['state'] == 'FAILURE':
                    return u'FAILURE'
            return st
        else:
            return u'NONE'

    def get_extended_state(self):
        """
        Returns the extended batch state.

        Raises:
            NidabaInputException if the batch hasn't been executed yet.
        """
        if not self.id:
            raise NidabaInputException('Object not attached to batch.')
        r = requests.get('{}/batch/{}'.format(self.host, self.id))
        r.raise_for_status()
        if 'chains' in r.json():
            self.lock = True
            return r.json()['chains']

    def get_results(self):
        """
        Retrieves the storage tuples of a successful batch.

        Returns:
        """
        if not self.id:
            raise NidabaInputException('Object not attached to batch.')
        r = requests.get('{}/batch/{}'.format(self.host, self.id))
        r.raise_for_status()
        if 'chains' in r.json():
            self.lock = True
            batch = r.json()['chains']
            outfiles = []
            for subtask in batch.itervalues():
                if len(subtask['children']) == 0 and not subtask['housekeeping'] and subtask['result'] is not None:
                    outfiles.append((subtask['result'], subtask['root_document']))
            return outfiles
        else:
            return None

    def get_tasks(self):
        """
        Returns the task tree either from the scratchpad or from the pipeline
        when already in execution.
        """
        if not self.id:
            raise NidabaInputException('Object not attached to batch.')
        r = requests.get('{}/batch/{}/tasks'.format(self.host, self.id))
        r.raise_for_status()
        return r.json()

    def get_documents(self):
        """
        Returns the list of input document for this task.
        """
        if not self.id:
            raise NidabaInputException('Object not attached to batch.')
        r = requests.get('{}/batch/{}/pages'.format(self.host, self.id))
        r.raise_for_status()
        return r.json()

    def add_document(self, path, callback, auxiliary=False):
        """
        Add a document to the batch.

        Uploads a document to the API server and adds it to the batch.

        ..note::
            Note that this function accepts a standard file system path and NOT
            a storage tuple as a client using the web API is not expected to
            keep a separate, local storage medium.

        Args:
            path (unicode): Path to the document
            callback (function): A function that is called with a
                                 ``requests_toolbelt.multipart.encoder.MultipartEncoderMonitor`` 
                                instance.
            auxiliary (bool): Switch to disable setting the file as an input
                              document. May be used to upload ground truths,
                              metadata, and other ancillary files..

        Raises:
            NidabaInputException: The document does not refer to a file or the
                                  batch is locked because the run() method has
                                  been called.
        """
        if not self.id:
            raise NidabaInputException('Object not attached to batch.')
        if self.lock:
            raise NidabaInputException('Executed batch may not be modified')
        if auxiliary:
            params = {'auxiliary': True}
        else:
            params = {}
        m = encoder.MultipartEncoderMonitor.from_fields(
            fields={'scans': (os.path.basename(path), open(path, 'rb'))},
            callback=callback)
        r = requests.post('{}/batch/{}/pages'.format(self.host, self.id),
                          data=m, headers={'Content-Type': m.content_type},
                          params=params)
        r.raise_for_status()
        return r.json()[0]['url']

    def add_task(self, group, method, *args, **kwargs):
        """
        Add a particular task configuration to a task group.

        Args:
            group (unicode): Group the task belongs to
            method (unicode): Name of the task
            kwargs: Arguments to the task
        """
        if not self.id:
            raise NidabaInputException('Object not attached to batch.')
        if self.lock:
            raise NidabaInputException('Executed batch may not be modified')
        # validate that the task exists
        if group not in self.allowed_tasks or method not in self.allowed_tasks[group]:
            raise NidabaInputException('Unknown task {}'.format(method))
        args = self.allowed_tasks[group][method]
        # validate against arg_values field of the task
        task_arg_validator(args, **kwargs)
        r = requests.post('{}/batch/{}/tasks/{}/{}'.format(self.host, self.id,
                                                           group, method),
                          json=kwargs)  
        r.raise_for_status()

    def run(self):
        """
        Executes the current batch definition.

        Expands the current batch definition to a series of celery chains and
        executes them asynchronously. Additionally a batch record is written to
        the celery result backend.

        Returns:
            (unicode): Batch identifier.
        """
        if not self.id:
            raise NidabaInputException('Object not attached to batch.')
        if self.lock:
            raise NidabaInputException('Executed batch may not be reexecuted')
        r = requests.post('{}/batch/{}'.format(self.host, self.id))
        r.raise_for_status()
        return self.id
